# -*- coding: utf-8 -*-

"""
Module pyelegantsdds.elegantrun 
=================================================================

A module containing the class ElegantRun to run Elegant simulations in 
a singularity container.

"""

import os
import shlex
import subprocess as subp
from io import StringIO

import numpy as np
import pandas as pd
from scipy import constants as const

from .elegant_command import ElegantCommandFile
from .sdds import SDDS, SDDSCommand

# ==============================================================================
#
# HELPER FUNCTIONS
#
# ==============================================================================


def write_parallel_elegant_script():
    """
    Method to generate a script that runs
    pelegant from bash.
    """

    # list of strings to write
    bashstrlist = [
        "#!/usr/bin/env bash",
        "if [ $# == 0 ] ; then",
        '   echo "usage: run_Pelegant <inputfile>"',
        "   exit 1",
        "fi",
        "n_cores=`grep processor /proc/cpuinfo | wc -l`",
        "echo The system has $n_cores cores.",
        "n_proc=$((n_cores-1))",
        "echo $n_proc processes will be started.",
        "if [ ! -e ~/.mpd.conf ]; then",
        '  echo "MPD_SECRETWORD=secretword" > ~/.mpd.conf',
        "  chmod 600 ~/.mpd.conf",
        "fi",
        "mpiexec -host $HOSTNAME -n $n_proc Pelegant  $1 $2 $3 $4 $5 $6 $7 $8 $9",
    ]

    bashstr = "\n".join(bashstrlist)

    # write to file
    with open("temp_run_pelegant.sh", "w") as f:
        f.write(bashstr)


def write_parallel_run_script(sif):
    """
    Method to generate parallel elegant run
    script.

    Parameters:
    -----------
    sif: str
                                                                                                                                                                                                                                                                    path to singularity container

    """
    bashstrlist = [
        "#!/bin/bash",
        "pele={}".format(sif),
        'cmd="bash temp_run_pelegant.sh"',
        "",
        "$pele $cmd $1",
    ]
    bashstr = "\n".join(bashstrlist)

    # write to file
    with open("run_pelegant.sh", "w") as f:
        f.write(bashstr)


def GenerateNDimCoordinateGrid(N, NPOINTS, pmin=1e-6, pmax=1e-4, man_ranges=None):
    """
    Method to generate an N dimensional coordinate grid for tracking,
    with fixed number of point in each dimension.
    The final shape is printed at creation.

    IMPORTANT:
    Number of grid points scales with N * NPOINTS**N, i.e.
    very large arrays are generated already with
    quite some small numbers for NPOINTS and N.

    Example: NPOINTS = 2, N = 6 -> 6*2*6 = 384 elements

    Parameters:
    -----------
    N: int
                                                                                                                                    dimension of the coordinate grid
    NPOINTS: int
                                                                                                                                    number of points in each dimension
    pmin: float
                                                                                                                                    min coordinate value in each dim
    pmax: float
                                                                                                                                    max coordinate value in each dim

    Returns:
    ========
    coordinate_grid : numpy array
                                                                                                                                    coordinate grid with particle ID in last column
    """
    rangelist = [np.linspace(pmin, pmax, NPOINTS)] * N
    if man_ranges is not None:
        # print(man_ranges)
        for k, v in man_ranges.items():
            rangelist[int(k)] = v
            # print(rangelist)
    grid = np.meshgrid(*rangelist)
    coordinate_grid = np.array([*grid])
    npart = coordinate_grid.size // N
    coordinate_grid = coordinate_grid.reshape(N, npart).T
    print("Shape: {} - Number of paritcles: {} ".format(coordinate_grid.shape, npart))
    # add particle id
    coordinate_grid = np.hstack((coordinate_grid, np.array(range(1, npart + 1)).reshape(npart, 1)))
    # print(coordinate_grid)

    return coordinate_grid


def generate_sphere_grid(dim=2, rmin=1e-6, rmax=1, rsteps=3, phisteps=3, **kwargs):
    """Method to generate grid point within n-dim ball, like polar but n-dim.
    Dimension 6 is a special case - as we need it for Elegant tracking. In this case
    the final two dimensions are not polar but fixed for dim 5 and in dim 6 and array
    passed via the kwargs 'deltaGamma'.

    Parameters
    ----------
    dim : int, optional
                                                                                                                                    dimension of the ball, by default 2
    rmin : float, optional
                                                                                                                                    minimal radius to use, by default 1e-6
    rmax : float, optional
                                                                                                                                    maximal radius to use, by default 1
    rsteps : int, optional
                                                                                                                                    number of steps in radius grid, by default 3
    phisteps : int, optional
                                                                                                                                    number of steps in the angle grid, by default 3
    """
    R = np.linspace(rmin, rmax, rsteps)
    mangle = np.pi

    # only track one kwadrant
    if kwargs.get("half", False):
        mangle = mangle / 2.0

    PHI1 = np.linspace(0, mangle, phisteps)
    PHI2 = np.linspace(
        0, mangle, phisteps
    )  # full sphere is 2 pi reduced for tracking to upper half

    # the special case
    if dim != 6:
        matrices = (R,) + tuple((PHI1 for _ in range(dim - 2))) + (PHI2,)
    else:
        # elegant t shift is fixed to zero
        # TODO: fix the fixed t shift
        matrices = (
            (R,)
            + tuple((PHI1 for _ in range(dim - 4)))
            + (PHI2,)
            + (np.array([0.0]), kwargs.get("deltaGamma", np.array([0.0])))
        )

    # create meshgrid to make all combinations
    meshmatrices = np.array(np.meshgrid(*matrices))

    # count the number of particles
    npart = meshmatrices.size // dim

    # reshape
    coord_T = meshmatrices.reshape(dim, npart).T

    #     X = (coord_T[:,0] * np.cos(coord_T[:,1]),)
    X = tuple()

    if dim == 6:
        ndim = 4
    else:
        ndim = dim

    for i in range(1, ndim):
        X += (coord_T[:, 0] * np.prod(np.sin(coord_T[:, 1:i]), axis=1) * np.cos(coord_T[:, i]),)

    X += (coord_T[:, 0] * np.prod(np.sin(coord_T[:, 1:-1]), axis=1) * np.sin(coord_T[:, -1]),)

    if dim != 6:
        sphere_grid = np.vstack(X)
    else:
        sphere_grid = np.vstack(X + (coord_T[:, 4], coord_T[:, 5]))
    print("Shape: {} - Number of paritcles: {} ".format(sphere_grid.T.shape, npart))

    # add particle id
    coordinate_grid = np.hstack((sphere_grid.T, np.array(range(1, npart + 1)).reshape(npart, 1)))
    # print(coordinate_grid)
    return coordinate_grid


# ==============================================================================
#
# MAIN CLASS
#
# ==============================================================================


class ElegantRun:
    """
    Class to interact with Elegant and Parallel Elegant from Python.
    """

    _REQUIRED_KWARGS = ["use_beamline", "energy"]

    def __init__(self, sif, lattice: str, parallel=False, **kwargs):
        self.sif = sif
        self.lattice = lattice
        self.parallel = parallel
        self.kwargs = kwargs
        self.check()
        self.commandfile = ElegantCommandFile("temp.ele")

        # setting up executable
        if parallel:
            self._write_parallel_script()
            self.exec = "bash {}".format(self.pelegant)
        else:
            self.exec = "{} elegant ".format(self.sif)

    def check(self):
        """
        Check if all necessary info
        is given for Elegant to be able to run.
        """
        if not all(req in self.kwargs.keys() for req in self._REQUIRED_KWARGS):
            print("Missing required kwargs...")
            print("Minimum required are:")
            for r in self._REQUIRED_KWARGS:
                print(r)

    def _write_parallel_script(self):
        """
        Generate the script to run parallel elegant.
        Method sets self.pelegant to the script file.
        """
        write_parallel_elegant_script()
        write_parallel_run_script(self.sif)
        self.pelegant = "run_pelegant.sh"

    def clearCommands(self):
        """Clear the command list."""
        self.commandfile.clear()

    def clearCommandHistory(self):
        """Clear the command history."""
        self.commandfile.clearHistory()

    def clearAll(self):
        """Clears both command list and command history."""
        self.clearCommands()
        self.clearCommandHistory()

    def run(self):
        """
        Run the commandfile.
        """
        # check if commandfile is not empty
        if len(self.commandfile.commandlist) == 0:
            print("Commandfile empty - nothing to do.")
            return

        # write Elegant command file to disk
        self.commandfile.write()

        # generate command string
        # Debug: print(self.exec)
        cmdstr = "{} temp.ele".format(self.exec)
        # Debug: print(cmdstr)

        # run
        with open(os.devnull, "w") as f:
            subp.call(shlex.split(cmdstr), stdout=f)

    def add_basic_setup(self, **kwargs):
        """
        Add basic setup command.
        """
        self.commandfile.clear()
        self.commandfile.addCommand(
            "run_setup",
            lattice=self.lattice,
            use_beamline=self.kwargs.get("use_beamline", None),
            p_central_mev=self.kwargs.get("energy", 1700.00),
            # centroid="%s.cen",
            default_order=kwargs.get("default_order", 2),
            concat_order=kwargs.get("concat_order", 1),
            rootname="temp",
            parameters="%s.params",
            semaphore_file="%s.done",
            magnets="%s.mag",  # for plotting profile
            final="%s.fin",
        )

    def add_basic_twiss(self):
        """
        Add basic twiss.
        """
        self.commandfile.addCommand(
            "twiss_output", filename="%s.twi", matched=1, radiation_integrals=1
        )

    def add_vary_element(self, **kwargs):
        """
        Add single vary element line.
        """
        self.commandfile.addCommand(
            "vary_element",
            name=kwargs.get("name", "*"),
            item=kwargs.get("item", "L"),
            intial=kwargs.get("initial", 0.0000),
            final=kwargs.get("final", 0.0000),
            index_number=kwargs.get("index_number", 0),
            index_limit=kwargs.get("index_limit", 1),
        )

    def add_vary_element_from_file(self, **kwargs):
        """
        Add single vary element line, loading value from
        dataset file.
        """
        if "enumeration_file" not in kwargs.keys():
            print("External filename missing.")
        else:
            self.commandfile.addCommand(
                "vary_element",
                name=kwargs.get("name", "*"),
                item=kwargs.get("item", "L"),
                index_number=kwargs.get("index_number", 0),
                index_limit=kwargs.get("index_limit", 1),
                enumeration_file=kwargs.get("enumeration_file"),
                enumeration_column=kwargs.get("enumeration_column"),
            )

    def add_basic_controls(self):
        """Adding basic controls for tracking"""
        # add controls
        self.commandfile.addCommand("run_control")
        self.commandfile.addCommand("bunched_beam")
        self.commandfile.addCommand("track")

    def add_watch(self, **kwargs):
        """Add watch point."""
        self.commandfile.addCommand(
            "insert_elements",
            name=kwargs.get("name", ""),
            type=kwargs.get("type", ""),
            exclude="",
            s_start=kwargs.get("s_start", -1),
            s_end=kwargs.get("s_end", -1),
            skip=kwargs.get("skip", 1),
            insert_before=kwargs.get("insert_before", 0),
            add_at_end=kwargs.get("add_at_end", 0),
            add_at_start=kwargs.get("add_at_start", 0),
            element_def=kwargs.get(
                "element_def",
                r'"WQ: WATCH, FILENAME=\"%s-%03ld.wq\", mode=\"coordinates\", interval={}"'.format(
                    kwargs.get("interval", 1)
                ),
            ),
        )

    def add_watch_at_start(self):
        """Add watch point at start of lattice."""
        self.add_watch(
            name="W",
            add_at_start=1,
            element_def=r'"W: WATCH, FILENAME=\"%s-%03ld.wq\", mode=\"coordinates\""',
        )

    def add_alter_elements(self, **kwargs):
        """Add alter_element command."""
        self.commandfile.addCommand("alter_elements", **kwargs)

    def add_rf_setup(self, **kwargs):
        """Add rf_setup command."""
        self.commandfile.addCommand(
            "rf_setup",
            filename="%s.rf",
            harmonic=kwargs.get("harmonic", 400),
            total_voltage=kwargs.get("total_voltage", 1.5e6),
        )

    def add_rf_get_freq_and_phase(
        self, total_voltage: float = 4 * 375e3, harmonic: int = 400, rad: bool = False, **kwargs
    ):
        """Add simulation single turn to get synchronuous freq and phase to be used in next simulation.
        Produces temp.twi and temp.out that can be used with tags.
        Parameters
        ----------
        total_voltage : float, optional
            Total sum of RF voltages, by default 4*375e3
        harmonic : int, optional
            harmonic number, by default 400
        rad : bool, optional
            radiation flag, by default False
        """

        self.commandfile.addCommand(
            "run_setup",
            lattice=self.lattice,
            use_beamline=self.kwargs.pop("use_beamline", None),
            p_central_mev=self.kwargs.pop("energy", 1700.00),
            # centroid="%s.cen",
            default_order=kwargs.pop("default_order", 1),
            concat_order=kwargs.pop("concat_order", 1),
            rootname="temp",
            parameters="%s.params",
            semaphore_file="%s.done",
            magnets="%s.mag",  # for plotting profile
            final="%s.fin",
            output="%s.out",
        )
        if rad:
            self.add_radiation_damping()

        self.add_basic_twiss()
        self.commandfile.addCommand(
            "rf_setup", total_voltage=total_voltage, harmonic=harmonic, track_for_frequency=1
        )
        self.add_basic_controls()
        self.run()

    def add_radiation_damping(self):
        """Add radiation damping for CSBEND, KQUAD and KSEXT."""

        self.add_alter_elements(name="*", type="CSBEND", item="SYNCH_RAD", value=1)
        self.add_alter_elements(name="*", type="CSBEND", item="USE_RAD_DIST", value=1)
        self.add_alter_elements(name="*", type="KQUAD", item="SYNCH_RAD", value=1)
        self.add_alter_elements(name="*", type="KQUAD", item="ISR", value=1)
        self.add_alter_elements(name="*", type="KQUAD", item="ISR1PART", value=1)
        self.add_alter_elements(name="*", type="KSEXT", item="SYNCH_RAD", value=1)
        self.add_alter_elements(name="*", type="KSEXT", item="ISR", value=1)
        self.add_alter_elements(name="*", type="KSEXT", item="ISR1PART", value=1)

    def use_standard_nkicks(self):
        self.add_alter_elements(name="*", type="CSBEND", item="N_KICKS", value=16)
        self.add_alter_elements(name="*", type="KQUAD", item="N_KICKS", value=8)
        self.add_alter_elements(name="*", type="KSEXT", item="N_KICKS", value=8)

    def add_fma_command(self, **kwargs):
        """
        Add elegant standard fma command.
        """

        self.commandfile.addCommand(
            "frequency_map",
            output="%s.fma",
            xmin=kwargs.get("xmin", -0.1),
            xmax=kwargs.get("xmax", 0.1),
            ymin=kwargs.get("ymin", 1e-6),
            ymax=kwargs.get("ymax", 0.1),
            delta_min=kwargs.get("delta_min", 0),
            delta_max=kwargs.get("delta_max", 0),
            nx=kwargs.get("nx", 21),
            ny=kwargs.get("ny", 21),
            ndelta=kwargs.get("ndelta", 1),
            verbosity=0,
            include_changes=kwargs.get("include_changes", 1),
            quadratic_spacing=kwargs.get("quadratic_spacing", 0),
            full_grid_output=kwargs.get("full_grid_output", 1),
        )

    def add_DA_command(self, **kwargs):
        """
        Add DA find aperture command.
        """
        self.commandfile.addCommand(
            "find_aperture",
            output="%s.aper",
            mode=kwargs.get("mode", "n-line"),
            verbosity=0,
            xmin=kwargs.get("xmin", -0.1),
            xmax=kwargs.get("xmax", 0.1),
            xpmin=kwargs.get("xpmin", 0.0),
            xpmax=kwargs.get("xpmax", 0.0),
            ymin=kwargs.get("ymin", 0.0),
            ymax=kwargs.get("ymax", 0.1),
            ypmin=kwargs.get("ypmin", 0.0),
            ypmax=kwargs.get("ypmax", 0.0),
            nx=kwargs.get("nx", 21),
            ny=kwargs.get("ny", 11),
            n_lines=kwargs.get("n_lines", 11),
            split_fraction=kwargs.get("split_fraction", 0.5),
            n_splits=kwargs.get("n_splits", 0),
            desired_resolution=kwargs.get("desired_resolution", 0.01),
            offset_by_orbit=kwargs.get("offset_by_orbit", 0),
            full_plane=kwargs.get("full_plane", 1),
        )

    def findtwiss(self, **kwargs):
        """
        Run Twiss and return Twiss parameters
        together with Twiss data.

        Parameters:
        ----------
        kwargs  : dict
                                                                                                                                        twiss command options
        """
        # TODO: add matched = 0 case
        matched = kwargs.get("matched", 1)
        initial_optics = kwargs.get("initial_optics", [])
        alternate_element = kwargs.get("alternate_elements", {})
        closed_orbit = kwargs.get("closed_orbit", 1)

        # make sure not residual is there
        self.commandfile.clear()

        # add setup command
        self.add_basic_setup()

        # add twiss calc
        self.commandfile.addCommand(
            "twiss_output",
            matched=matched,
            output_at_each_step=0,
            filename="%s.twi",
            radiation_integrals=1,
        )

        # add controls
        self.add_basic_controls()

        # write command file
        self.commandfile.write()

        # set cmdstr and run
        cmdstr = "{} elegant temp.ele".format(self.sif)
        with open(os.devnull, "w") as f:
            subp.call(shlex.split(cmdstr), stdout=f)

        # load twiss output
        twifile = SDDS(self.sif, "temp.twi", 0)
        twiparams = twifile.getParameterValues()
        twidata = twifile.getColumnValues()

        twiparams["length"] = np.round(twidata.iloc[-1]["s"], 3)

        return twidata, twiparams

    def find_matrices(self, **kwargs):
        """
        Find element by element matrix and map elements (depending on given order).
        Constant vector and R matrix are returned as numpy arrays, the maps are
        returned as dicts.

        Parameters:
        -----------
        kwargs  :
                                                                                                                                        - SDDS_output_order : order of maps (max is 3)

        Returns:
        --------
        C       : np.array
                                                                                                                                        constant vector
        R       : np.array
                                                                                                                                        R matrix
        T_dict  : dict
                                                                                                                                        T map Tijk as key
        Q_dict  : dict
                                                                                                                                        U map Qijkl as key
        """
        assert kwargs.get("SDDS_output_order", 1) < 4

        self.commandfile.clear()
        self.add_basic_setup()
        self.commandfile.addCommand(
            "matrix_output",
            SDDS_output="%s.sdds",
            SDDS_output_order=kwargs.get("SDDS_output_order", 1),
            printout="%s.mat",
            printout_order=kwargs.get("SDDS_output_order", 1),
            full_matrix_only=kwargs.get("full_matrix_only", 0),
            individual_matrices=kwargs.get("individual_matrices", 1),
            output_at_each_step=kwargs.get("output_at_each_step", 1),
        )

        # add controls
        self.add_basic_controls()

        # write command file
        self.commandfile.write()

        # set cmdstr
        cmdstr = "{} elegant temp.ele".format(self.sif)
        with open(os.devnull, "w") as f:
            subp.call(shlex.split(cmdstr), stdout=f)

        with open("temp.mat", "r") as f:
            mdata = f.read()

        # get full turn matrix and
        dfmat = pd.read_csv(
            StringIO("\n".join(mdata.split("full", 1)[1].splitlines()[1:])),
            delim_whitespace=True,
            names=[1, 2, 3, 4, 5, 6],
        )
        C = dfmat.loc[dfmat.index == "C:"].values.T
        R = dfmat.loc[dfmat.index.str.contains("R")].values
        T = dfmat.loc[dfmat.index.str.contains("T")]
        Q = dfmat.loc[dfmat.index.str.contains("Q")]

        T_dict = {}
        for _, row in T.iterrows():
            _basekey = row.name[:-1]
            for c in T.columns:
                _key = _basekey + str(c)
                _value = row[c]
                if not pd.isna(_value):
                    T_dict[_key] = _value

        Q_dict = {}
        for _, row in Q.iterrows():
            _basekey = row.name[:-1]
            for c in Q.columns:
                _key = _basekey + str(c)
                _value = row[c]
                if not pd.isna(_value):
                    Q_dict[_key] = _value

        sddsmat = SDDS(self.sif, "temp.sdds", 0)
        ElementMatrices = sddsmat.getColumnValues()

        return C, R, ElementMatrices, T_dict, Q_dict

    def generate_sdds_particle_inputfile(self, grid_type="rectangular", **kwargs):
        """
        Generates an SDDS file containing initial
        particle coordinates on a grid. The grid
        can be defined through the kwargs.

        Parameters:
        ----------
        kwargs      :
                                        - pmin: min value of grid on each dim
                                        - pmax: max value of grid on each dim
                                        - pcentralmev: particle energy (code converts it to beta * gamma )
                                        - man_ranges: dict containing as key dim num - in order x xp y yp s p
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        and as values an array of values to be used
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        For p this is autoset to beta gamma based on pcentralmev
                                        - NPOINTS: number of linear spaced points in each dim for the grid

        Returns:
        --------
        None, writes the data to pre-defined named file.

        """
        assert grid_type in ["rectangular", "spherical"]
        pcentral = kwargs.get("pcentralmev", self.kwargs.get("energy"))
        print(pcentral)
        # convert to beta * gamma
        pcentral = np.sqrt(
            (pcentral / const.physical_constants["electron mass energy equivalent in MeV"][0]) ** 2
            - 1
        )

        if grid_type == "rectangular":
            npoints_per_dim = kwargs.get("NPOINTS", 2)
            pmin = kwargs.get("pmin", 0)
            pmax = kwargs.get("pmax", 1e-4)
            man_ranges = kwargs.get("man_ranges", {"5": np.array([pcentral])})
            if "5" not in man_ranges.keys() and 5 not in man_ranges.keys():
                man_ranges["5"] = np.array([pcentral])
            # example : man_ranges={'0':np.array([1e-6,1e-5]),'1':[0]})

            # generate coordinate grid, with particle id as last column
            # and save it as plain data table seperated by a whitespace
            particle_df = pd.DataFrame(
                GenerateNDimCoordinateGrid(
                    6, npoints_per_dim, pmin=pmin, pmax=pmax, man_ranges=man_ranges
                )
            )
            particle_df.to_csv("temp_plain_particles.dat", sep=" ", header=None, index=False)

            # cleanup kwargs
            kwargs.pop("NPOINTS", None)
            kwargs.pop("pmin", None)
            kwargs.pop("pmax", None)
            kwargs.pop("man_ranges", None)
        else:
            rmin = kwargs.get("rmin", 1e-6)
            rmax = kwargs.get("rmax", 1e-1)
            rsteps = kwargs.get("rsteps", 3)
            half = kwargs.get("half", True)
            phisteps = kwargs.get("phisteps", 5)
            deltaGamma = kwargs.get("deltaGamma", np.array([pcentral]))

            particle_df = pd.DataFrame(
                generate_sphere_grid(
                    dim=6,
                    rmin=rmin,
                    rmax=rmax,
                    rsteps=rsteps,
                    phisteps=phisteps,
                    deltaGamma=deltaGamma,
                    half=half,
                )
            )

            particle_df.to_csv("temp_plain_particles.dat", sep=" ", header=None, index=False)
            # clean up kwargs
            kwargs.pop("rmin", None)
            kwargs.pop("rmax", None)
            kwargs.pop("rsteps", None)
            kwargs.pop("half", None)
            kwargs.pop("phisteps", None)
            kwargs.pop("deltaGamma", None)

        kwargs.pop("pcentralmev", None)

        # Create sddscommand object
        sddscommand = SDDSCommand(self.sif)

        # update the command parameters
        if self.parallel:
            outputmode = "binary"
        else:
            outputmode = "ascii"
        kwargs["outputMode"] = outputmode
        kwargs["file_2"] = (
            "temp_particles_input.txt" if not self.parallel else "temp_particles_input.bin"
        )

        # load the pre-defined  convert plain data to sdds command
        cmd = sddscommand.get_particles_plain_2_SDDS_command(**kwargs)

        # run the sdds command
        sddscommand.runCommand(cmd)

        self.sdds_beam_file = kwargs["file_2"]

    def simple_single_particle_track(self, coord=np.zeros((5, 1)), **kwargs):
        """
        Track a single particle with given initial coordinates.

        Important:
        ----------
        Be careful with giving the 6th coordinate, this is beta * gamma. If not
        given it will be calculated automatically either using standard 1700 MeV
        or kwargs["pcentralmev"].

        """
        # generate particle input file
        self.generate_sdds_particle_inputfile(
            man_ranges={k: v for k, v in zip(range(coord.shape[0] + 1), coord)}, **kwargs
        )

        # construct command file
        self.commandfile.clear()
        self.add_basic_setup()

        self.commandfile.addCommand("run_control", n_passes=kwargs.get("n_passes", 2 ** 8))
        self.commandfile.addCommand("bunched_beam")
        self.commandfile.addCommand(
            "sdds_beam",
            input=self.sdds_beam_file,
            input_type='"elegant"',
        )
        self.commandfile.addCommand("track")

        # run will write command file and execute it
        self.run()

    def track_simple(self, **kwargs):
        """
        Track a set of particles.
        """
        # construct command file
        self.commandfile.clear()
        self.add_basic_setup()
        if kwargs.get("add_watch_start", False):
            self.add_watch_at_start()

        if kwargs.get("rad", False):
            self.add_radiation_damping()

        if kwargs.get("rf", False):
            self.add_basic_twiss()
            #            # get frequency
            #            self.commandfile.addCommand("rpn_load", filename="%s.out", tag="ref", use_row=0)
            #
            #            # get energy loss
            #            self.add_basic_twiss()
            #            self.commandfile.addCommand(
            #                "rpn_load", filename="%s.twi", tag="reftwi", load_parameters=1
            #            )
            #
            #            # update RF settings
            #            self.add_alter_elements(
            #                name="*", type="RFCA", item="VOLT", value=kwargs.get("total_voltage") / 4.0
            #            )
            #            self.add_alter_elements(
            #                name="*", type="RFCA", item="FREQ", value='("ref.t rec 400 * ")'
            #            )
            #            self.add_alter_elements(
            #                name="*",
            #                type="RFCA",
            #                item="PHASE",
            #                value='("180 reftwi.U0 1e6 * {} / dasin -")'.format(kwargs.get("total_voltage")),
            #            )
            #
            #            self.add_alter_elements(
            #                name="*",
            #                type="RFCA",
            #                item="PHASE",
            #                value='("180")'.format(kwargs.get("total_voltage")),
            #            )
            #
            self.add_rf_setup(
                filename="%s.rf", harmonic=400, total_voltage=kwargs.get("total_voltage")
            )

        self.commandfile.addCommand("run_control", n_passes=kwargs.get("n_passes", 2 ** 8))
        self.commandfile.addCommand("bunched_beam")
        self.commandfile.addCommand(
            "sdds_beam",
            input=self.sdds_beam_file,
            input_type='"elegant"',
        )

        self.commandfile.addCommand("track")

        # run will write command file and execute it
        self.run()

    def track_vary(
        self, varydict: dict, varyitemlist=None, mode="row", add_watch_start=False, **kwargs
    ):
        """
        Track a set of particles in combination with a
        vary command.
        """
        assert varyitemlist is not None
        assert len(varyitemlist) == len(varydict)
        assert mode.lower() in ["row", "table"]

        # generate the sdds input file
        sdds = SDDS(self.sif, "temp.sdds", 0)
        sdds.generate_scan_dataset(varydict)

        self.commandfile.clear()
        self.add_basic_setup()

        if kwargs.get("rf", False):
            self.add_rf(volt=kwargs.get("volt", 350000))

        if kwargs.get("rad", False):
            self.add_radiation_damping()

        if add_watch_start:
            self.add_watch_at_start()
        n_idx = 1 if mode == "row" else len(varydict)
        self.commandfile.addCommand(
            "run_control", n_indices=n_idx, n_passes=kwargs.get("n_passes", 2 ** 8)
        )
        if mode == "table":
            for i, it in enumerate(varydict.items()):
                k, v = it
                self.add_vary_element_from_file(
                    name=k,
                    item=varyitemlist[i],
                    index_number=i,
                    index_limit=len(v),
                    enumeration_file="temp.sdds",
                    enumeration_column=k,
                )
        else:
            for i, it in enumerate(varydict.items()):
                k, v = it
                self.add_vary_element_from_file(
                    name=k,
                    item=varyitemlist[i],
                    index_number=0,
                    index_limit=len(v),
                    enumeration_file="temp.sdds",
                    enumeration_column=k,
                )
        self.commandfile.addCommand("bunched_beam")
        self.commandfile.addCommand(
            "sdds_beam", input=self.sdds_beam_file, input_type='"elegant"', reuse_bunch=1
        )
        self.commandfile.addCommand("track")
        self.run()

    def fma(self, **kwargs):
        """
        Run Elegant fma.
        """
        self.commandfile.clear()
        self.add_basic_setup()

        self.commandfile.addCommand("run_control", n_passes=kwargs.pop("n_passes", 2 ** 8))
        self.add_basic_twiss()
        self.add_fma_command(**kwargs)

        self.run()

    def dynap(self, **kwargs):
        """
        Run Elegant's Dynamic Aperture.
        """
        self.commandfile.clear()
        self.commandfile.addCommand(
            "run_setup",
            lattice=self.lattice,
            use_beamline=self.kwargs.get("use_beamline", None),
            p_central_mev=self.kwargs.get("energy", 1700.00),
            centroid="%s.cen",
            default_order=kwargs.get("default_order", 2),
            concat_order=kwargs.get("concat_order", 3),
            rootname="temp",
            parameters="%s.params",
            semaphore_file="%s.done",
            magnets="%s.mag",  # for plotting profile
            losses="%s.los",
        )

        self.commandfile.addCommand("twiss_output", filename="%s.twi", output_at_each_step=1)
        self.commandfile.addCommand("run_control", n_passes=kwargs.pop("n_passes", 2 ** 9))
        self.add_DA_command(**kwargs)

        self.run()

    def dynapmom(self):
        """
        Run Elegant's Dynamic Momentum Aperture.
        """
        # TODO
        pass

    def table_scan(self, scan_list_of_dicts, mode="row", add_watch_start=True, **kwargs):
        """ """
        assert mode in ["row", "table"]
        n_idx = 1 if mode == "row" else len(scan_list_of_dicts)
        self.commandfile.clear()
        self.add_basic_setup()

        if add_watch_start:
            self.add_watch_at_start()
        self.commandfile.addCommand(
            "run_control", n_indices=n_idx, n_passes=kwargs.get("n_passes", 2 ** 8)
        )
        for i, l in enumerate(scan_list_of_dicts):
            if mode == "table":
                inx = i
            else:
                inx = 0
            self.commandfile.addCommand(
                "vary_element",
                name=l.get("name"),
                item=l.get("item"),
                initial=l.get("initial"),
                final=l.get("final"),
                index_number=inx,
                index_limit=l.get("index_limit"),
            )
        self.commandfile.addCommand(
            "sdds_beam", input=self.sdds_beam_file, input_type='"elegant"', reuse_bunch=1
        )
        self.commandfile.addCommand("track")

        # run will write command file and execute it
        self.run()



# -*- coding: utf-8 -*-

"""
Module pyelegantsdds.sdds 
=================================================================

A module containing the SDDSCommand class and  SDDS class.

The SDDSCommand class is responsible for creating, editing, etc ... of SDDS
commands.

The SDDS class uses the SDDSCommand class to build template and composed commands.

"""

import os
import shlex
import subprocess as subp
from io import StringIO
from itertools import count
from typing import Dict, List

import numpy as np
import pandas as pd
from dask import dataframe as dd

from .tools.sddsutils import sddsconvert2ascii, sddsconvert2binary


# ================================================================================
# SDDSCommand class:
#
#
# ================================================================================
class SDDSCommand:
    """SDDSCommand manupilation class"""

    _COMMANDLIST = [
        "plaindata2sdds",
        "sdds2stream",
        "sddsconvert",
        "sddsquery",
        "sddsprocess",
        "sddsplot",
        "sddsprintout",
        "sddsmakedataset",
        "sddsnaff",
        "sddsresdiag",
    ]

    def __init__(self, sif):
        self.sif = sif
        self.command = {}

    def createCommand(self, command: str, note: str = "", **params) -> None:
        """
        Method to add a command to the command file.
        Generates a dict to reconstruct command string,
        that can be added to the command file.

        Parameters:
        ----------
        command     : str
        note        : str

        Key Parameters:
        ---------------
        params : dict

        Returns:
        --------
        None
        """
        # check if it is a valid Elegant command
        if self.checkCommand(command) == False:
            print("The command {} is not recognized.".format(command))
            return

        # init command dict
        thiscom = {}
        thiscom["NAME"] = command

        # add command parameters to the command dict
        for k, v in params.items():
            thiscom[k] = v

        # add the command dict to the command list
        self.command = thiscom

    def checkCommand(self, typename: str) -> bool:
        """
        Check if a command is a valid
        Elegant command.

        Parameters:
        -----------
        typename    : str

        Returns:
        --------
        boolean

        """
        for tn in self._COMMANDLIST:
            if tn == typename.lower():
                return True
        return False

    def clearCommand(self):
        """
        Clear the command list.
        """
        self.command = {}

    def getCommand(self, command, **params):
        """Method to get the command.

        Parameters
        ----------
        command : dict

        Returns
        -------
        str
        command string
        """
        # create command
        self.createCommand(command, **params)

        # init command string
        cmdstr = "{} {}".format(self.sif, self.command["NAME"])

        # build the command string
        for k, v in params.items():
            if "file" in k.lower():
                cmdstr += " {}".format(v)
            elif "separator" in k.lower():
                cmdstr += ' "-separator= {}"'.format(v)
            else:
                if v is not None:
                    cmdstr += " -{}={}".format(k.split("_")[0], v)
                else:
                    cmdstr += " -{}".format(k.split("_")[0])

        return cmdstr

    def runCommand(self, commandstring=None):
        """Run the command.

        Parameters
        ----------
        commandstring : str
        """
        if commandstring is None:
            name = self.command.pop("NAME")
            commandstring = self.getCommand(name, **self.command)
        print("Running command {}".format(commandstring))
        subp.run(commandstring, check=True, shell=True)

    def get_particles_plain_2_SDDS_command(self, **params) -> str:
        """
        Returns sdds command to turn plain data table, separated
        by a *space* into SDDS particle initial coordinates SDDS file.
        Easy to generate plain data with pandas.

        """
        return self.getCommand(
            "plaindata2sdds",
            file_1=params.get("file_1", "temp_plain_particles.dat"),
            file_2=params.get("file_2", "temp_particles_input.txt"),
            inputMode=params.get("inputMode", "ascii"),
            outputMode=params.get("outputMode", "ascii"),
            separator=" ",
            column_1="x,double,units=m",
            column_2="xp,double",
            column_3="y,double,units=m",
            column_4="yp,double",
            column_5="t,double,units=s",
            column_6='p,double,units="m$be$nc"',
            columns_7="particleID,long",
            noRowCount=None,
        )


# ================================================================================
# SDDS CLASS
# ================================================================================


class SDDS:
    """
    Class for interacting with SDDS files.
    """

    def __init__(self, sif: str, filename: str, filetype: int):
        assert filetype in [0, 1]
        self.sif = sif
        self._filetype = filetype
        self._filename = filename
        self.columnlist = None
        self.parameterlist = None
        self.commandlist = []
        self.command_history = {}
        self._count_iter = count(start=0, step=1)

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, value):
        self._filename = value

    @property
    def filetype(self):
        return self._filetype

    @filetype.setter
    def filetype(self, value):
        assert value in [0, 1]
        self._filetype = value

    def addCommand(self, command: str, **params) -> None:
        """Add a command to the command list.

        Parameters
        ----------
        command : str
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        valid sdds command
        """
        sddscommand = SDDSCommand(self.sif)
        cmdstr = sddscommand.getCommand(command, **params)

        # add command to current commandlist
        self.commandlist.append(cmdstr)

    def clearCommandList(self, save: bool = True) -> None:
        """
        Clear the command list.

        Parameters:
        -----------
        bool: save
        """
        if save and len(self.commandlist) > 0:
            self._addHistory(self.commandlist)

        self.commandlist = []

    def clearHistory(self):
        """
        Clear command history.
        """
        self.command_history = {}
        self._count_iter = count(start=0, step=1)

    def _addHistory(self, cmdlist):
        """
        Add commands to history.

        Parameters:
        -----------
        list | str : cmdlist
        """
        _key = next(self._count_iter)
        if not isinstance(cmdlist, list):
            cmdlist = [cmdlist]
        _value = cmdlist
        self.command_history[_key] = _value

    def runCommand(self):
        """
        Run all the sddscommands in the command list.
        """
        # print info
        if len(self.commandlist) > 0:
            # add commands to history
            self._addHistory(self.commandlist)

            for cmd in self.commandlist:
                # run command
                print("Executing : \n{}".format(cmd))

                p = subp.Popen(cmd, stdout=subp.PIPE, shell=True)
                (output, err) = p.communicate()
                p_status = p.wait()

        else:
            print("No commands entered - nothing to do!")

        # clear the commandlist
        self.clearCommandList()

    def printHistory(self):
        """Print command history."""
        if len(self.command_history) > 0:
            for i, l in self.command_history.items():
                print("History key: {}\n".format(i))
                print("---------------\n")
                for c in l:
                    print("{}\n".format(c))

                # print extra newline
                print("\n")
        else:
            print("History is empty.")

    def reload_from_history(self, history_idx: int) -> None:
        """Reloads the commandlist from history with index
        history_idx.

        Parameters
        ----------
        history_idx : int
                                                                        index of history command to reload
                                                                        use `printHistory()` to get an overview
        """
        self.clearCommandList(save=True)
        self.commandlist = self.command_history.get(history_idx)

    def rerun_from_history(self, history_idx):
        """Rerun a history entry.count()

        Parameters
        ----------
        history_idx : int
        """
        assert history_idx in list(self.command_history.keys())

        self.clearCommandList(save=True)
        self.commandlist = self.command_history.get(history_idx)
        self.runCommand()

    def load_raw_data(self):
        """Read the data from file."""
        if self.filetype == 1:
            # ASCII FORMAT
            with open(self.filename, "r") as f:
                self.raw_content = f.read()

        else:
            # BINARY FORMAT
            with open(self.filename, "rb") as f:
                self.raw_content = f.read()

    def convert(self, outfile: str = None):
        """Convert between filetypes.

        Parameters
        ----------
        outfile : str
        """
        if self.filetype == 0:
            converted_filename = self.filename + ".txt"

            if outfile is not None:
                converted_filename = outfile

            cmdstr = "{} sddsconvert -ascii {} {}".format(
                self.sif, self.filename, converted_filename
            )
        else:
            converted_filename = self.filename + ".bin"

            if outfile is not None:
                converted_filename = outfile

            cmdstr = "{} sddsconvert -binary {} {}".format(
                self.sif, self.filename, converted_filename
            )
        # add to command history
        self._addHistory(cmdstr)

        # reset filename
        print("Warning - auto filename set")
        print("Changed from {} to {}".format(self.filename, converted_filename))
        self.filename = converted_filename

        # reset filetype
        print("Warning - auto filetype set")
        print("Changed from {} to {}".format(self.filetype, abs(1 - self.filetype)))
        self.filetype = abs(1 - self.filetype)

        with open(os.devnull, "w") as f:
            subp.call(shlex.split(cmdstr), stdout=f)

    def getColumnList(self):
        """Get a list of columns available in sdds file.

        Returns
        -------
        list
        """
        # command string
        cmdstr = "{} sddsquery -columnList {}".format(self.sif, self.filename)

        # add to command history
        self._addHistory(cmdstr)

        # run the command
        p = subp.Popen(cmdstr, stdout=subp.PIPE, shell=True)
        (output, err) = p.communicate()
        p_status = p.wait()

        # set the columnlist
        self.columnlist = [l.decode("utf-8") for l in output.splitlines()]

        return self.columnlist

    def getColumnValues(self, memory_threshold=100e6):
        """Get column values from file and return either
        as dataframe or return filename of converted data.

        Parameters
        ----------
        memory_threshold : float, optional

        Returns
        -------
        str | pandas.DataFrame
        """
        # make sure we have the column names
        if self.columnlist is None:
            self.getColumnList()

        # data can be large (bigger than memory)
        # use threshold to preve
        if os.path.getsize(self.filename) > memory_threshold:
            print(
                "File is large, output redirected to file {}".format(
                    self.filename + "_columnvalues.dat"
                )
            )

            # build command string
            cmdstr = "{} sdds2stream -col={} {} > {}".format(
                self.sif,
                ",".join(self.columnlist),
                self.filename,
                self.filename + "_columnvalues.dat",
            )

            # run command
            p = subp.Popen(cmdstr, stdout=subp.PIPE, shell=True)
            (output, err) = p.communicate()
            p_status = p.wait()

            return self.filename + "_columnvalues.dat"

        else:
            # build command string
            cmdstr = "{} sdds2stream -col={} {}".format(
                self.sif, ",".join(self.columnlist), self.filename
            )

            # run command
            p = subp.Popen(cmdstr, stdout=subp.PIPE, shell=True)
            (output, err) = p.communicate()
            p_status = p.wait()

            # load file output into dataframe
            output = output.decode("utf-8")
            output = pd.read_csv(StringIO(output), names=self.columnlist, delim_whitespace=True)

            return output

    def getParameterList(self):
        """Get list of parameter names available in sdds file.

        Returns
        -------
        list
        """
        # command string
        cmdstr = "{} sddsquery -parameterList {}".format(self.sif, self.filename)

        # add to command history
        self._addHistory(cmdstr)

        # run command
        p = subp.Popen(cmdstr, stdout=subp.PIPE, shell=True)
        (output, err) = p.communicate()
        p_status = p.wait()

        # set the parameter list
        self.parameterlist = [l.decode("utf-8") for l in output.splitlines()]

        return self.parameterlist

    def getParameterValues(self):
        """Get parameter values from sdds file.

        Returns
        -------
        dataframe
        """
        cmdstr = "{} sddsprintout -parameters=* -spreadsheet {}".format(self.sif, self.filename)
        p = subp.Popen(cmdstr, stdout=subp.PIPE, shell=True)
        (output, err) = p.communicate()
        p_status = p.wait()
        df = pd.read_csv(
            StringIO(output.decode("utf-8")),
            error_bad_lines=False,
            delim_whitespace=True,
            skip_blank_lines=True,
            skiprows=1,
            names=["ParameterName", "ParameterValue"],
            index_col=False,
        )

        df = df.set_index("ParameterName", drop=True)
        df = pd.to_numeric(
            df.drop(["SVNVersion", "Stage", "PreviousElementName"], errors="ignore")[
                "ParameterValue"
            ]
        )
        self.ParameterName = df
        return df

    def readParticleData(self, vary=False, long=False):
        """Read the particle tracking data.

        Parameters
        ----------
        vary : bool, optional
                                                                        adds a step column for distinguising between the
                                                                        different vary_element steps.
        Returns
        -------
        DataFrame (pandas or dask)
        """
        # if vary command was used in ele command file
        # convert data and add step column
        if vary:
            self.process_scan()

        if long:
            self.addCommand(
                "sddsprocess",
                define='column,delta,"p pCentral / 1 - " {}'.format(self.filename),
                outfile="{}_longitudinal.{}".format(
                    self.filename.split(".")[0], self.filename.split(".")[1]
                ),
            )

            # run command
            self.runCommand()

            # cmdstr = "{} sddsprocess -define=column,step,Step {} {}_processed.{}".format(
            #    self.sif, self.filename, *self.filename.split(".")
            # )
            # print(cmdstr)
            # p = subp.Popen(cmdstr, stdout=subp.PIPE, shell=True)
            newfilename = "{}_longitudinal.{}".format(*self.filename.split("."))
            print("Warning - auto filename set")
            print("Changed from {} to {}".format(self.filename, newfilename))

            # update the filename
            self.filename = newfilename

        self.getColumnList()

        # if the file type is binary convert it first
        if self.filetype == 0:
            self.convert(outfile=None)

        # try to load the column data
        df = self.getColumnValues()

        # check if data or filename is returned
        if isinstance(df, str):
            # if filename was returned lazy load the data using dask dataframe
            data = dd.read_csv(df, delimiter=" ", names=self.columnlist, header=None)
        else:
            data = df

        # if vary commmand was used group the data and add Turn column
        if vary:
            grouped = data.groupby(by="step")

            def f(group):
                return group.join(
                    pd.DataFrame({"Turn": group.groupby("particleID").cumcount() + 1})
                )

            data = grouped.apply(f)

        else:
            data["Turn"] = data.groupby("particleID").cumcount() + 1

        # return data eager or lazy depending on the datasize
        return data

    def process_scan(self):
        """Process sdds file containig particle tracking data where
        *vary_element* was used in the command file/
        """
        # add command
        self.addCommand(
            "sddsprocess",
            define="column,step,Step {}".format(self.filename),
            outfile="{}_processed.{}".format(
                self.filename.split(".")[0], self.filename.split(".")[1]
            ),
        )

        # run command
        self.runCommand()

        # cmdstr = "{} sddsprocess -define=column,step,Step {} {}_processed.{}".format(
        #    self.sif, self.filename, *self.filename.split(".")
        # )
        # print(cmdstr)
        # p = subp.Popen(cmdstr, stdout=subp.PIPE, shell=True)
        newfilename = "{}_processed.{}".format(*self.filename.split("."))
        print("Warning - auto filename set")
        print("Changed from {} to {}".format(self.filename, newfilename))

        # update the filename
        self.filename = newfilename

    def sddsplot_tunediagram(self, **kwargs):

        self.addCommand("sddsresdiag", file1="resdiag.sdds")
        self.runCommand()

        newkwargs = {
            "columnNames": "nux,nuy",
            "file": self.filename,
            "scale": "0,1,0,1",
            "graph": "sym,fill,vary=subtype",
            "order": "spect",
            "split": "col=x",
            "col": "nux,nuy",
            "file2": "resdiag.sdds",
            "sever": None,
        }
        newkwargs = {**newkwargs, **kwargs}
        self.addCommand("sddsplot", **newkwargs)

        self.runCommand()

    def sddsplot_base(self, **kwargs):
        """
        Basic sddsplot command.
        """
        sddscommand = SDDSCommand(self.sif)
        cmd = sddscommand.getCommand("sddsplot", file=self.filename, **kwargs)

        # add to command history
        self._addHistory(cmd)

        sddscommand.runCommand(cmd)

    def sddsplot(
        self,
        file=None,
        columnNames="x,xp",
        scale="0,0,0,0",
        graph="dot,vary",
        order="spectral",
        split="columnBin=particleID",
        **kwargs,
    ):
        """
        Quick standard multi particle tracking plotting
        of phase space - add columnNames to plot other
        variables.

        Colors are assigned per particle ID.

        Parameters:
        -----------
        """
        if file is not None:
            newkwargs = {
                "file": file,
                "columnNames": columnNames,
                "scale": scale,
                "graph": graph,
                "order": order,
                "split": split,
            }

            newkwargs = {**newkwargs, **kwargs}
            self.sddsplot_base(**newkwargs)
        else:
            print("File missing.")

    def sddsplot_fma(
        self,
        file=None,
        col="x,y",
        split="column=diffusionRate",
        graph="sym,vary=subtype,fill,scale=2,fill",
        order="spectral",
        **kwargs,
    ):
        newkwargs = {
            "file": file,
            "col": col,
            "graph": graph,
            "order": order,
            "split": split,
        }
        if file is None:
            newkwargs.pop("file")
        newkwargs = {**newkwargs, **kwargs}
        self.sddsplot_base(**newkwargs)

    def generate_scan_dataset(self, datasetdict):
        """
		Generates a file called "temp_scan.sdds" containing columns of values
		to be used by elegant to scan over using vary_element method.

		Arguments:
		----------
		datadict: dict
			dictionary where the keys are the column headers and values are list of values to scan over
			Note: all dict values need to have the same length
		
		Example:
		--------
		>>> datasetdc = {
			"Q1" : [1.205055,1.555550505],
			"Q2" : [-1.45000,-1.800000000],\
			"Q3D": [-2.02000,-2.020000000],\
			"Q4D": [1.408000,1.4080000000],\
			"Q3T": [0.000000,0.0000000000],\
			"Q4T": [0.000000,0.0000000000],\
			"Q5T": [0.000000,0.0000000000],\
			"S1" : [0.000000,0.0000000000],\
			"S2" : [0.000000,0.0000000000],\
			"S3D": [0.000000,0.0000000000],\
			"S4D": [0.000000,0.0000000000],\
			"S3T": [0.000000,0.0000000000],\
			"S4T": [0.000000,0.0000000000]
		}
		>>> sdds = SDDS(sif, "temp.sdds",0)
		>>> sdds.generate_scan_dataset(datasetdc)

		"""
        cmd = f"{self.sif}  sddsmakedataset {self.filename} "

        for k, v in datasetdict.items():
            cmd += f"-column={k},type=double -data=" + ",".join([str(vv) for vv in v]) + " "

        subp.run(cmd, check=True, shell=True)
        self.sdds_scan = "temp.sdds"



# -*- coding: utf-8 -*-

"""
Module pyelegantsdds.tools.sddsutils
=================================================================

A module containing a list of sdds comamnds.

"""
import os
import subprocess
import time
from itertools import product

import numpy as np
import pandas as pd
from dask import dataframe as dd
from termcolor import colored


def sddsconvert2ascii(sif, filename):
    """
    Convert sdds binary file to ascii.

    Parameters:
    -----------
    sif: str
        Singularity executable container where sdds is installed.

    filename: str
        filename of the file to convert

    Returns:
    --------
    None
        A new file is created "filename.txt" in ascii format.
    """
    subprocess.run(f"{sif} sddsconvert -ascii {filename} {filename}.txt", shell=True)


def sddsconvert2binary(sif, filename):
    """
    Convert sdds ascii file to binary.

    Parameters:
    -----------
    sif: str
        Singularity executable container where sdds is installed.

    filename: str
        filename of the file to convert

    Returns:
    --------
    None
        A new file is created "filename.txt" in binary format.

    """
    subprocess.run(f"{sif} sddsconvert -binary {filename} {filename}.bin", shell=True)


def getParameterList(parameterlistfile):
    """
    Method to read back the parameter list,
    extracted with sddsextractparameternames.

    Parameters:
    -----------
    parameterlistfile: str
        Filename of the file where the parameter list is
        stored.

    Returns:
    --------
    parameterlist: List[str]
        List of the parameter names.
    """
    with open(parameterlistfile, "r") as f:
        lines = f.read().splitlines()

    return [line for line in lines]


def sddsextractparameternames(sif, elefma, ext="-001.w1"):
    """
    Extract the parameter names from an SDDS file.

    Parameters:
    -----------
    sif: str
        Singularity executable container where sdds is installed.
    elefma: str
        Elegant .ele base name (filename of the .ele file without the extension)
    ext: str
        Extension of the file, format is "elefma{ext}", usually watchpoint extension as
        in default.

    Returns:
    --------
    filename: str
        Filename where the parameter list is stored.
    parameterlist: List[str]
        List of extracted parameter names.

    """
    subprocess.run(
        f"{sif} sddsquery  -parameterList " f"{elefma}{ext} > {elefma}_parameterlist.txt",
        shell=True,
    )

    filename = f"{elefma}_parameterlist.txt"

    return filename, getParameterList(filename)


def getColumnList(columnlistfile):
    """
    Method to read back the column list,
    extracted with sddsextractcolumnnames.

    Parameters:
    -----------
    columnlistfild: str
        Filename of the file where the column list is
        stored.

    Returns:
    --------
    columnlist: List[str]
        List of the column names.
    """
    with open(columnlistfile, "r") as f:
        lines = f.read().splitlines()

    return [line for line in lines]


def sddsextractcolumnnames(sif, elefma, ext="-001.w1"):
    """
    Extract the columns names from an SDDS file.

    Parameters:
    -----------
    sif: str
        Singularity executable container where sdds is installed.
    elefma: str
        Elegant .ele base name (filename of the .ele file without the extension)
    ext: str
        Extension of the file, format is "elefma{ext}", usually watchpoint extension as
        in default.

    Returns:
    --------
    filename: str
        Filename where the column list is stored.
    columnlist: List[str]
        List of extracted column names.

    """
    subprocess.run(
        f"{sif} sddsquery -columnList " f"{elefma}{ext} > {elefma}_columnlist.txt", shell=True
    )

    filename = f"{elefma}_columnlist.txt"

    return filename, getColumnList(filename)


def sddsextractparametervalues():
    """
    Extract the parameter values from an sdds file.
    :return:
    """
    pass


def sddsextractcolumnvalues(sif, elefma, cols="x,xp,y,yp,t,p,dt,particleID", ext="-001.w1"):
    """
    Extract the column values from an sdds file.

    Parameters:
    -----------
    sif: str
        Singularity executable container where sdds is installed.
    elefma: str
        Elegant .ele base name (filename of the .ele file without the extension)
    cols: str
        String representation of the list of columns to extract the values for.
    ext: str
        Extension of the file, format is "elefma{ext}", usually watchpoint extension as
        in default.

    Returns:
    --------
    filename: str
        Filename where the values are stored.

    """
    subprocess.run(
        f"{sif} sdds2stream -col={cols} " f"{elefma}{ext} > {elefma}_particle_data.txt", shell=True
    )
    return f"{elefma}_particle_data.txt"


def processvaryelementoutput(sif, basename_binary):
    """
    Processes coordinate output file from tracking in combination
    with vary element. It adds a column "step" such that the resulting
    coordinate table can be grouped per "step" and "particleID".
    This allows to study individual tracked particles for individual vary
    steps.

    Parameters:
    -----------
    sif : str
        Singularity executable container where sdds is installed.add()
    basename_binary : str
        name of the file to be processed

    Returns:
    --------
    filename : str
        Filename where the processed data is stored.

    """
    subprocess.run(
        f"{sif} sddsprocess -define=column,step,Step {basename_binary} processed_{basename_binary}",
        check=True,
        shell=True,
    )
    return f"processed_{basename_binary}"


def generate_scan_dataset(sif, datasetdict, filepath):
    """
    Generates a file called "scan.sdds" containing columns of values
    to be used by elegant to scan over using vary_element method.

    Parameters:
    -----------
    datadict: dict
        dictionary where the keys are the column headers and values are list of values to scan over
        Note: all dict values need to have the same length

    filepath: str
        path where the simulation will be run (i.e where ele and lte files are)

    Returns:
    --------
    None
        Creates file scan.sdds in the filepath to be used in the simulations.add()

    """
    # get current working dir to be able to get back
    currdir = os.getcwd()

    # change to simulation dir
    os.chdir(filepath)
    print("File path used: {}".format(filepath))

    # create scan.sdds
    cmd = f"{sif}  sddsmakedataset scan.sdds "

    for k, v in datasetdict.items():
        cmd += f"-column={k},type=double -data=" + ",".join([str(vv) for vv in v]) + " "

    subprocess.run(cmd, check=True, shell=True)

    # change back to original working dir
    os.chdir(currdir)


def sddsplot(
    sif,
    filepath,
    columnNames=["x", "xp"],
    markerstyle="sym",
    vary="subtype",
    scalemarker=1,
    fill=True,
    order="spectral",
    split="page",
    scale="0,0,0,0",
):
    """
    Method to generate sdds plot.

    Parameters:
    -----------
    sif:
    filepath
    columnNames
    markerstyle
    vary
    scalemarker
    """
    if fill:
        strfill = ",fill"
    else:
        strfill = ""
    cmd = f"{sif} sddsplot -columnNames={','.join(columnNames)} {filepath} "
    cmd += f"-graph={markerstyle},vary={vary}{strfill},scale={str(scalemarker)} -order={order} -split={split} -scale={scale}"
    subprocess.run(cmd, check=True, shell=True)


def readParticleData(coordfile, collist, simtype="spt", vary=False):
    """ """
    data = dd.read_csv(coordfile, delimiter=" ", names=collist, header=None)
    if vary:
        grouped = data.groupby(by="step")

        def f(group):
            return group.join(pd.DataFrame({"Turn": group.groupby("particleID").cumcount() + 1}))

        data = grouped.apply(f)

    elif simtype == "spt":
        data["Turn"] = data.groupby("particleID").cumcount() + 1

    return data



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyelegantsdds', 'pyelegantsdds.tools']

package_data = \
{'': ['*']}

install_requires = \
['PyNAFF>=1.1.4,<2.0.0',
 'dask>=2021.4.1,<2022.0.0',
 'ipykernel>=5.5.4,<6.0.0',
 'ipython>=7.23.1,<8.0.0',
 'isort>=5.8.0,<6.0.0',
 'matplotlib>=3.4.2,<4.0.0',
 'nbsphinx>=0.8.4,<0.9.0',
 'numpy>=1.20.2,<2.0.0',
 'pandas>=1.2.4,<2.0.0',
 'pandoc>=1.0.2,<2.0.0',
 'scipy>=1.6.3,<2.0.0',
 'termcolor>=1.1.0,<2.0.0',
 'tqdm>=4.60.0,<5.0.0']

setup_kwargs = {
    'name': 'pyelegantsdds',
    'version': '0.2.0',
    'description': '<Enter a one-sentence description of this project here.>',
    'long_description': '=============\nPyElegantSDDS\n=============\n\n\nPyElegantSDDS is a Python wrapper around a Singularity container with SDDS and Parallel\nElegant installed in it. It uses this container to build and run templates for common task in \nparticle tracking for accelerators.\n\n* Free software: MIT license\n* Documentation: https://PyElegantSDDS.readthedocs.io.\n\n\nFeatures\n--------\n\n* SDDS commands\n* Elegant commands (twiss, track, FMA, DA, ...)\n* Run simulations in just a few lines of Python code. \n* Generate Hyper-cubical and Hyper-Spherical distributions for tracking\n* Quick plotting using SDDS tools\n\n',
    'author': 'Tom Mertens',
    'author_email': 'your.email@whatev.er',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/tomerten/PyElegantSDDS',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-

"""
Module pyelegantsdds.elegant_command 
=================================================================

A module containing the ElegantCommandFile class.

"""

from itertools import count


class ElegantCommandFile:
    _COMMANDLIST = (
        "alter_elements",
        "amplification_factors",
        "analyze_map",
        "aperture_data",
        "bunched_beam",
        "change_particle",
        "chromaticity",
        "closed_orbit",
        "correct",
        "correction_matrix_output",
        "correct_tunes",
        "coupled_twiss_output",
        "divide_elements",
        "error_element",
        "error_control",
        "find_aperture",
        "floor_coordinates",
        "frequency_map",
        "global_settings",
        "insert_elements",
        "insert_sceffects",
        "linear_chromatic_tracking_setup",
        "link_control",
        "link_elements",
        "load_parameters",
        "matrix_output",
        "modulate_elements",
        "moments_output",
        "momentum_aperture",
        "optimize",
        "optimization_constraint",
        "optimization_covariable",
        "optimization_setup",
        "parallel_optimization_setup",
        "optimization_term",
        "optimization_variable",
        "print_dictionary",
        "ramp_elements",
        "rf_setup",
        "replace_elements",
        "rpn_expression",
        "rpn_load",
        "run_control",
        "run_setup",
        "sasefel",
        "save_lattice",
        "sdds_beam",
        "semaphores",
        "slice_analysis",
        "subprocess",
        "steering_element",
        "touschek_scatter",
        "transmute_elements",
        "twiss_analysis",
        "twiss_output",
        "track",
        "tune_shift_with_amplitude",
        "vary_element",
    )

    def __init__(self, filename):
        self.commandlist = []
        self.filename = filename
        self.history = {}
        self._count_iter = count(start=0, step=1)

    def checkCommand(self, typename):
        """
        Check if a command is a valid
        Elegant command.

        Parameters:
        ----------
        typename    : str
                command type

        """
        for tn in self._COMMANDLIST:
            if tn == typename.lower():
                return True
        return False

    def addCommand(self, command, note="", **params):
        """
        Method to add a command to the command file.
        Generates a dict to reconstruct command string,
        that can be added to the command file.

        Parameters:
        ----------
        command     : str
                valid Elegant command
        note        : str
                extra info
        """
        # check if it is a valid Elegant command
        if self.checkCommand(command) == False:
            print("The command {} is not recognized.".format(command))
            return

        # init command dict
        thiscom = {}

        thiscom["NAME"] = command.lower()
        thiscom["NOTE"] = note

        # add command parameters to the command dict
        for k, v in params.items():
            if v != "":
                thiscom[k] = v

        # add the command dict to the command list
        self.commandlist.append(thiscom)

    def modifyCommand(self, commandname, mode="last", **params):
        """
        Method to modify a command already in the command list.
        If the command is used multiple times, the command you
        want to change can be select by using the strings "first" or "last"
        or by giving the integer index of the command.

        Parameters:
        ----------
        commandname     : str
                name of the command to modify
        mode            : str | int, default last
                if multiple commands select which one to change
        params          : dict
                command parameters - updated

        """
        # create command index list to be able
        # to choose which one to change
        indlist = []
        i = 0

        for command in self.commandlist:
            if command["NAME"] == commandname.lower():
                indlist.append(i)
            i += 1

        # find command to updated given in mode
        # allowed are : first, last or integer
        if len(indlist) == 1:

            for k, v in params.items():
                self.commandlist[indlist[0]][k] = v
            return

        elif len(indlist) == 0:
            print("No such command {} can be found".format(commandname))

        else:
            if mode == "last":
                for k, v in params.items():
                    self.commandlist[indlist[-1]][k] = v
                return
            elif mode == "first":
                for k, v in params.items():
                    self.commandlist[indlist[0]][k] = v
                return

            elif isinstance(mode, int):
                for k, v in params.items():
                    self.commandlist[indlist[mode]][k] = v
                return

            else:
                print("The mode is invalid")

    def repeatCommand(self, commandname, mode="last"):
        """
        Method to repeat a commend already in the
        command list.


        Parameters:
        ----------
        commandname     : str
                name of the command to modify
        mode            : str | int , default last
                if multiple commands select which one to change

        """
        # create index list for the command to be repeated
        indlist = []
        i = 0
        for command in self.commandlist:
            if command["NAME"] == commandname.lower():
                indlist.append(i)
            i += 1

        if len(indlist) == 1:
            self.commandlist.append(self.commandlist[indlist[0]])
            return

        elif len(indlist) == 0:
            print("No such command {} can be found".format(commandname))

        else:
            if mode == "last":
                self.commandlist.append(self.commandlist[indlist[-1]])
            elif mode == "first":
                self.commandlist.append(self.commandlist[indlist[0]])
            elif isinstance(mode, int):
                self.commandlist.append(self.commandlist[indlist[mode]])
            else:
                print("The mode is invalid")

    def clearHistory(self):
        """
        Clear command history.
        """
        self.history = {}
        self._count_iter = count(start=0, step=1)

    def _addHistory(self, cmdlist):
        """
        Add commands to history.

        Parameters:
        -----------
        list | str : cmdlist
                command or list of commands to add to history.
        """
        _key = next(self._count_iter)
        if not isinstance(cmdlist, list):
            cmdlist = [cmdlist]
        _value = cmdlist
        self.history[_key] = _value

    def clear(self):
        """Clear command list."""
        # check if commadnlist empty if not add to history
        if len(self.commandlist) > 0:
            self._addHistory(self.commandlist)

        # clear commandlist
        self.commandlist = []

    def remove_command(self, commandname, mode="last"):
        # create command index list to be able
        # to choose which one to change
        clist = [d.get("NAME") for d in self.commandlist]
        indlist = [i for i, x in enumerate(clist) if x == commandname]

        if mode == "last":
            self.commandlist = (
                self.commandlist[: indlist[-1]] + self.commandlist[indlist[-1] + 1 :]
            )
        elif mode == "first":
            self.commandlist = self.commandlist[: indlist[0]] + self.commandlist[indlist[0] + 1 :]
        else:
            self.commandlist = (
                self.commandlist[: indlist[mode]] + self.commandlist[indlist[mode] + 1 :]
            )

    def write(self, outputfilename="", mode="w"):
        """
        Method to write the command file to external file.

        Parameters:
        -----------
        outputfilename: str
                filename to write to
        mode : str
                python file write mode
        """
        if outputfilename != "":
            self.filename = outputfilename

        # writing the command file
        with open(self.filename, mode) as outfile:
            # looping over the commands
            for command in range(len(self.commandlist)):
                outfile.write("&{}\n".format(self.commandlist[command]["NAME"]))

                if self.commandlist[command]["NOTE"] != "":
                    outfile.write("! {}\n".format(self.commandlist[command]["NOTE"]))

                for k, v in self.commandlist[command].items():
                    if k != "NAME" and k != "NOTE":
                        if len(k) <= 19:
                            outfile.write("\t{:20s}= {},\n".format(k, v))
                        elif len(k) <= 29:
                            outfile.write("\t{:30s}= {},\n".format(k, v))
                        else:
                            outfile.write("\t{:40s}= {},\n".format(k, v))
                outfile.write("&end\n\n")
        self._addHistory(self.commandlist)
        self.clear()
